/* This is free and unencumbered software released into the public domain.
 *
 * Anyone is free to copy, modify, publish, use, compile, sell, or
 * distribute this software, either in source code form or as a compiled
 * binary, for any purpose, commercial or non-commercial, and by any
 * means.
 *
 * See LICENSE for complete information.
 */

#include <string.h>

#include "glp.h"
#include "glp_rand_openssl_aes.h"
#include "glp_utils.h"


/***********************************************SIGNING KEY GENERATION********************************/

/*generates signing key (s1,s2), stored in physical form */
void glp_gen_sk(glp_signing_key_t *sk){    
    sample_glp_secret(sk->s1);
    sample_glp_secret(sk->s2);
}

void glp_mask_sk(glp_masked_signing_key_t *msk, glp_signing_key_t* sk) {
    mask_poly(msk->s1, sk->s1);
    mask_poly(msk->s2, sk->s2);
}


/***********************************************PUBLIC KEY GENERATION********************************/

/*takes a signing key stored in physical space and computes the public key in physical space */
/*points a1, a2 are stored in FFT space */
void glp_gen_pk(glp_public_key_t *pk, glp_signing_key_t sk){
  FFT_FORWARD(sk.s1);
  FFT_FORWARD(sk.s2);
  POINTWISE_MUL_ADD(pk->t, a, sk.s1, sk.s2, N, Q);
  FFT_BACKWARD(sk.s1);
  FFT_BACKWARD(sk.s2);
  FFT_BACKWARD(pk->t);
}


/**********************************************SIGNING*********************************************/

#if NISPOWEROFTWO
#define N_LIMIT N
#else
#define N_LIMIT N-1
#endif

#define MAXSIGNITERS 1000

/*signs a message as (z,c) where z is a ring elt in physical form, and c is a hash output encoded as a sparse poly */
int glp_sign(glp_signature_t *sig,
              const glp_signing_key_t sk, 
              const unsigned char *message_digest,
              const size_t dgst_len){
  RINGELT y1[N]={0}, y2[N] ={0};
  int success = 0;
  uint16_t i, sign_iters = 0;
 
  /*initialise random sampling for y1,y2*/
  RANDOM_VARS
  /*sample y1,y2 randomly, and repeat until they pass rejection sampling*/
  while(!success && (sign_iters < MAXSIGNITERS)){
    sign_iters += 1;
#if !NISPOWEROFTWO
    y1[N-1] = 0;
    y2[N-1] = 0;
#endif
  for(i = 0; i < N_LIMIT; i++){
    while(1){
      y1[i] = RANDOM32; /*get 32 bits of random */
      y1[i] &= ~(~0 << (B_BITS + 1)); /*take bottom (B_BITS + 1) bits */
      if(y1[i] <= 2*B + 1) break;
    }
    while(1){
      y2[i] = RANDOM32; /*get 32 bits of random */
      y2[i] &= ~(~0 << (B_BITS + 1)); /*take bottom (B_BITS + 1) bits */
      if(y2[i] <= 2*B + 1)   break;
    }    
    y1[i] = (y1[i] <= B)? y1[i] : Q - (y1[i] - B); 
    y2[i] = (y2[i] <= B)? y2[i] : Q - (y2[i] - B); 
  }
  success = glp_deterministic_sign(sig, y1,y2, sk, message_digest, dgst_len);
  }
  return success;
}

int glp_masked_sign(glp_signature_t *sig,
              const glp_masked_signing_key_t msk, 
              const unsigned char *message_digest,
              const size_t dgst_len){
  MRINGELT y1[N], y2[N];
  int success = 0;
  uint16_t i, sign_iters = 0;
 
  /*sample y1,y2 randomly, and repeat until they pass rejection sampling*/
  while(!success && (sign_iters < MAXSIGNITERS)){
    sign_iters += 1;
    for(i = 0; i < N_LIMIT; i++){
	bint32_randomness_gen(y1[i]);
	bint32_randomness_gen(y2[i]);
    }
    success = glp_masked_deterministic_sign(sig,
	        y1,y2, msk, message_digest, dgst_len);
  }
  return success;
}

/*signs a message for a fixed choice of ephemeral secret y in physcial space
 returns 1 or 0 according to success or failure in doing so (due to rejection sampling)*/
int glp_deterministic_sign(glp_signature_t *signature, 
                           const RINGELT y1[N],
                           const RINGELT y2[N],
                           const glp_signing_key_t sk, 
                           const unsigned char *message_digest,
                           const size_t dgst_len){
  RINGELT u[N],u_rounded[N], y1_fft[N], y2_fft[N];
  unsigned char hash_output[GLP_DIGEST_LENGTH];
  copy_poly(y1_fft,y1);
  copy_poly(y2_fft,y2);
  FFT_FORWARD(y1_fft);
  FFT_FORWARD(y2_fft);

  /*u = a y1 + y2*/
  POINTWISE_MUL_ADD(u,a,y1_fft,y2_fft,N,Q);
  FFT_BACKWARD(u);
  MAPTOCYCLOTOMIC(u,N,Q);

  /*round and hash u*/
  copy_poly(u_rounded,u);
  round_poly(u_rounded,B - OMEGA);
  if(!hash(hash_output, u_rounded, message_digest,dgst_len)) return 0;
  if(!encode_sparse(&(signature->c), hash_output)) return 0;

  /*z_1 = y_1 + s_1 c*/
  sparse_mul(signature->z1, sk.s1, signature->c);
  POINTWISE_ADD(signature->z1,signature->z1,y1,N,Q);
  MAPTOCYCLOTOMIC(signature->z1,N,Q);

  /*rejection sampling on z_1*/
  for(uint16_t i = 0; i < N_LIMIT; i++) if(ABS(signature->z1[i]) > (B - OMEGA) ) return 0;

  /*z_2 = y_2 + s_2 c*/
  sparse_mul(signature->z2, sk.s2, signature->c);
  POINTWISE_ADD(signature->z2,signature->z2,y2,N,Q);
  MAPTOCYCLOTOMIC(signature->z2,N,Q);

  /*rejection sampling on z_2*/
  for(uint16_t i = 0; i < N_LIMIT; i++) if(ABS(signature->z2[i]) > (B - OMEGA) ) return 0;

  /*compression rounding_target = a*z1 - t*c = u - z2*/
  RINGELT approx_u[N], rounding_target[N];
  POINTWISE_SUB(rounding_target,u,signature->z2, N, Q);
  MAPTOCYCLOTOMIC(rounding_target,N,Q);

  copy_poly(approx_u,rounding_target);
  round_poly(approx_u,B -OMEGA);

  /*signature compression*/
  for(uint16_t i = 0; i < N_LIMIT; i++){
    if(approx_u[i] == u_rounded[i]) signature->z2[i] = 0;

    else if(rounding_target[i] <= (B-OMEGA)){
      if(2*(Q%(2*(B-OMEGA) + 1)) >= (B-OMEGA)){
        if((SIGN(signature->z2[i]) < 0) && (SIGN((rounding_target[i] + signature->z2[i]) %Q) <0)) signature->z2[i] = NEG(B-OMEGA);
        else signature->z2[i] = B-OMEGA;
      }
    }

    else if(rounding_target[i] >= Q - (B-OMEGA)){
      if((SIGN(signature->z2[i]) > 0) && (rounding_target[i] + signature->z2[i] >= Q))signature->z2[i] = (B-OMEGA);
      else if(approx_u[i] == u_rounded[i] + 1) signature->z2[i] = NEG(B-OMEGA);
    }

    else if(approx_u[i] == u_rounded[i] + 1) signature->z2[i] = NEG(B-OMEGA);
    else signature->z2[i] = B-OMEGA;
  }
  return 1;
}

int glp_masked_deterministic_sign(glp_signature_t *signature, 
                           MRINGELT y1[N],
                           MRINGELT y2[N],
                           glp_masked_signing_key_t sk, 
                           const unsigned char *message_digest,
                           const size_t dgst_len){
  MRINGELT u[N], z1[N], z2[N], ay1[N], mul[2*N];
  RINGELT u_unmasked[N],u_rounded[N];
  unsigned char hash_output[GLP_DIGEST_LENGTH];

  static RINGELT a0[N] = { 0 };
  if(a0[0] == 0) {
      copy_poly(a0, a);
      FFT_BACKWARD(a0);
  }

  int j,k;
  bint32_t tmp;
  for(j=0; j<2*N; j++) {
      bint32_to_base(mul[j], 0);
  }
  for(j=0; j<N; j++) {
      for(k=0; k<N; k++) {
	  bint32_mul_const_modp(tmp, y1[j], a0[k]);
	  bint32_add_arith_modp(mul[j+k], mul[j+k], tmp);
      }
  }

#if NISPOWEROFTWO
  /*reduce mod x^n + 1*/
  for(k = 0; k < N; k++)
    bint32_sub_arith_modp(ay1[k], mul[k], mul[k+N]);
#else
  /*reduce mod x^n - 1*/
  for(k = 0; k < N; k++)
    bint32_add_arith_modp(ay1[k], mul[k], mul[k+N]);
#endif

  for(k = 0; k < N; k++)
    bint32_add_arith_modp(u[k], ay1[k], y2[k]);

  refresh_unmask_poly(u_unmasked, u);

#if 0
  RINGELT u_test[N],y1_fft[N], y2_fft[N];

  refresh_unmask_poly(y1_fft,y1);
  refresh_unmask_poly(y2_fft,y2);
  FFT_FORWARD(y1_fft);
  FFT_FORWARD(y2_fft);
  POINTWISE_MUL_ADD(u_test,a,y1_fft,y2_fft,N,Q);
  FFT_BACKWARD(u_test);

  for(k = 0; k < N; k++)
    if(u_test[k] != u_unmasked[k]) {
      printf("Error: u_test and u_unmasked not equal!\n");
      return 0;
    }
#endif

  /*round and hash u*/
  copy_poly(u_rounded,u_unmasked);
  round_poly(u_rounded,B - OMEGA);
  if(!hash(hash_output, u_rounded, message_digest,dgst_len)) return 0;
  if(!encode_sparse(&(signature->c), hash_output)) return 0;

  /*z_1 = y_1 + s_1 c*/
  sparse_mul_masked(z1, sk.s1, signature->c);
  for(k = 0; k < N; k++)
    bint32_add_arith_modp(z1[k], z1[k], y1[k]);

#if 0
  RINGELT y1_unmasked[N], z1_unmasked[N], s1_unmasked[N];
  refresh_unmask_poly(y1_unmasked,y1);
  refresh_unmask_poly(z1_unmasked,z1);
  refresh_unmask_poly(s1_unmasked,sk.s1);
  sparse_mul(signature->z1, s1_unmasked, signature->c);
  POINTWISE_ADD(signature->z1,signature->z1,y1_unmasked,N,Q);

  for(k = 0; k < N; k++)
    if((signature->z1)[k] != z1_unmasked[k]) {
      printf("Error: signature->z1 and z1_unmasked not equal!\n");
      printf("z1_unmasked:");
      print_poly(z1_unmasked);
      printf("signature->z1:");
      print_poly(signature->z1);
      return 0;
    }
#endif

  /*z_2 = y_2 + s_2 c*/
  sparse_mul_masked(z2, sk.s2, signature->c);
  for(k = 0; k < N; k++)
    bint32_add_arith_modp(z2[k], z2[k], y2[k]);

  /*rejection sampling on z_2*/
  if(rejection_sampling(z1, z2))
    return 0;

  refresh_unmask_poly(signature->z1, z1);
  refresh_unmask_poly(signature->z2, z2);

  /*compression rounding_target = a*z1 - t*c = u - z2*/
  RINGELT approx_u[N], rounding_target[N];
  POINTWISE_SUB(rounding_target,u_unmasked,signature->z2, N, Q);
  MAPTOCYCLOTOMIC(rounding_target,N,Q);

  copy_poly(approx_u,rounding_target);
  round_poly(approx_u,B -OMEGA);

  /*signature compression*/
  for(uint16_t i = 0; i < N_LIMIT; i++){
    if(approx_u[i] == u_rounded[i]) signature->z2[i] = 0;

    else if(rounding_target[i] <= (B-OMEGA)){
      if(2*(Q%(2*(B-OMEGA) + 1)) >= (B-OMEGA)){
        if((SIGN(signature->z2[i]) < 0) && (SIGN((rounding_target[i] + signature->z2[i]) %Q) <0)) signature->z2[i] = NEG(B-OMEGA);
        else signature->z2[i] = B-OMEGA;
      }
    }

    else if(rounding_target[i] >= Q - (B-OMEGA)){
      if((SIGN(signature->z2[i]) > 0) && (rounding_target[i] + signature->z2[i] >= Q))signature->z2[i] = (B-OMEGA);
      else if(approx_u[i] == u_rounded[i] + 1) signature->z2[i] = NEG(B-OMEGA);
    }

    else if(approx_u[i] == u_rounded[i] + 1) signature->z2[i] = NEG(B-OMEGA);
    else signature->z2[i] = B-OMEGA;
  }
  return 1;
}

int rejection_sampling(MRINGELT z1[N], MRINGELT z2[N]) {
  MRINGELT coeffs[4*N];
  uint16_t j;

#if 0
  uint32_t b, truebit = 1;

  int32_t x;
  for(j = 0; j < N; j++) {
    x = bint32_recombine_modp_signed(z1[j]);
    if( (x > B-OMEGA) || (x < OMEGA-B) )
	truebit = 0;
      
    x = bint32_recombine_modp_signed(z2[j]);
    if( (x > B-OMEGA) || (x < OMEGA-B) )
	truebit = 0;
  }
#endif

  for(j = 0; j < N; j++) {
      bint32_copy(coeffs[j], z1[j]);
      bint32_copy(coeffs[j+N], z2[j]);
      bint32_neg_modp(coeffs[j+2*N], z1[j]);
      bint32_neg_modp(coeffs[j+3*N], z2[j]);
  }
#if 0
  b = bint32_all_large_enough(coeffs, 4*N);
  if(truebit != b) {
      printf("Error in rejsamp!\n");
  }
  return !b;
#endif
  return !bint32_all_large_enough(coeffs, 4*N);
}

/**********************************************VERIFICATION*********************************************/

int glp_verify(glp_signature_t sig, 
               const glp_public_key_t pk, 
               const unsigned char *message_digest,
               const size_t dgst_len){
  RINGELT u[N],v[N];  
  sparse_poly_t c_test;
  unsigned char hash_output[GLP_DIGEST_LENGTH];
  uint16_t i;
  for(i = 0; i < N_LIMIT; i++){
    if(ABS(sig.z1[i]) > (B - OMEGA) ) return 0;
    if(ABS(sig.z2[i]) > (B - OMEGA) ) return 0;
  }
  FFT_FORWARD(sig.z1);
  FFT_FORWARD(sig.z2);
  POINTWISE_MUL_ADD(u,a,sig.z1,sig.z2,N,Q);
  FFT_BACKWARD(u);
  sparse_mul(v,pk.t,sig.c);
  POINTWISE_SUB(u,u,v,N,Q);
  MAPTOCYCLOTOMIC(u,N,Q);
  round_poly(u, B-OMEGA);
  if(!hash(hash_output, u, message_digest, dgst_len))return 0;
  if(!encode_sparse(&c_test,hash_output))return 0;
  for(i = 0; i < OMEGA; i++){
    if(c_test.pos[i] != sig.c.pos[i]) return 0;
    if(c_test.sign[i] != sig.c.sign[i])return 0;
  }
    
  return 1;
}



